package net.astah.stpa.stamp.edit.sequencer;

import static java.util.Collections.synchronizedMap;
import static net.astah.stpa.stamp.STAMPPackage.Literals.ACCIDENT;
import static net.astah.stpa.stamp.STAMPPackage.Literals.ACCIDENT__HAZARD;
import static net.astah.stpa.stamp.STAMPPackage.Literals.CONTROL_ACTION;
import static net.astah.stpa.stamp.STAMPPackage.Literals.COUNTERMEASURE;
import static net.astah.stpa.stamp.STAMPPackage.Literals.HAZARD;
import static net.astah.stpa.stamp.STAMPPackage.Literals.HAZARD_CAUSAL_FACTOR;
import static net.astah.stpa.stamp.STAMPPackage.Literals.HAZARD_CAUSAL_FACTOR__COUNTERMEASURE;
import static net.astah.stpa.stamp.STAMPPackage.Literals.HAZARD_CAUSAL_FACTOR__UNSAFE_CONTROL_ACTION;
import static net.astah.stpa.stamp.STAMPPackage.Literals.HAZARD__SAFETY_CONSTRAINT;
import static net.astah.stpa.stamp.STAMPPackage.Literals.IDENTIFIED_ELEMENT__ID;
import static net.astah.stpa.stamp.STAMPPackage.Literals.PRECONDITION;
import static net.astah.stpa.stamp.STAMPPackage.Literals.SAFETY_CONSTRAINT;
import static net.astah.stpa.stamp.STAMPPackage.Literals.STPA_ANALYSIS__ACCIDENT;
import static net.astah.stpa.stamp.STAMPPackage.Literals.STPA_ANALYSIS__HAZARD_CAUSAL_FACTOR;
import static net.astah.stpa.stamp.STAMPPackage.Literals.STPA_ANALYSIS__PRECONDITION;
import static net.astah.stpa.stamp.STAMPPackage.Literals.UCA_ANALYSIS_RESULT__GUIDE_WORD;
import static net.astah.stpa.stamp.STAMPPackage.Literals.UCA_ANALYSIS_RESULT__UNSAFE_CONTROL_ACTION;
import static net.astah.stpa.stamp.STAMPPackage.Literals.UCA_ANALYSIS__CONTROL_ACTION;
import static net.astah.stpa.stamp.STAMPPackage.Literals.UCA_GUIDE_WORD;
import static net.astah.stpa.stamp.STAMPPackage.Literals.UCA_GUIDE_WORD__UCA_ID_POSTFIX;
import static net.astah.stpa.stamp.STAMPPackage.Literals.UNSAFE_CONTROL_ACTION;
import static net.astah.stpa.stamp.STAMPPackage.Literals.UNSAFE_CONTROL_ACTION__CONTROL_ACTION;

import java.util.List;
import java.util.Map;
import java.util.WeakHashMap;
import java.util.regex.Pattern;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.transaction.ResourceSetChangeEvent;
import org.eclipse.emf.transaction.ResourceSetListenerImpl;
import org.eclipse.emf.transaction.RollbackException;
import org.eclipse.emf.transaction.TransactionalEditingDomain;

import net.astah.emf.common.ecore.EComparators;
import net.astah.emf.common.notify.EFeaturePath;
import net.astah.emf.edit.sequencer.AttributeSequencer;
import net.astah.emf.edit.sequencer.IdReorganizer;
import net.astah.emf.edit.sequencer.IdSequencer;
import net.astah.emf.edit.sequencer.IdSequencerRegistryImpl;
import net.astah.emf.edit.sequencer.PatternSequencer;
import net.astah.emf.edit.sequencer.ReferenceIndexSequencer;

/**
 * トランザクションの完了を監視して、モデルの ID を再編成します。
 */
public class IdReorganizerListener extends ResourceSetListenerImpl {

    /**
     * {@code TransactionalEditingDomain} と {@code IdReorganizer} の対応関係を維持します。
     */
    protected final Map<TransactionalEditingDomain, IdReorganizer> reorganizers = synchronizedMap(
            new WeakHashMap<>());

    @Override
    public boolean isPrecommitOnly() {
        return true;
    }

    @Override
    public Command transactionAboutToCommit(final ResourceSetChangeEvent event)
            throws RollbackException {
        final TransactionalEditingDomain domain = event.getEditingDomain();
        final IdReorganizer reorganizer = reorganizers.get(domain);
        if (reorganizer != null) {
            final List<Notification> notifications = event.getNotifications();
            return reorganizer.createCommand(domain, notifications);
        }
        return null;
    }

    @Override
    public void setTarget(final TransactionalEditingDomain domain) {
        super.setTarget(domain);
        reorganizers.computeIfAbsent(domain, key -> createReorganizer());
    }

    @Override
    public void unsetTarget(final TransactionalEditingDomain domain) {
        super.unsetTarget(domain);
        final IdReorganizer reorganizer = reorganizers.remove(domain);
        if (reorganizer != null) {
            reorganizer.dispose();
        }
    }

    /**
     * STAMP Workbench の ID 自動採番ルールが定義された {@code IdReorganizer} を構築して返します。
     * 
     * @return 構築した {@code IdReorganizer}
     */
    protected IdReorganizer createReorganizer() {
        final IdSequencer.Registry registry = new IdSequencerRegistryImpl();

        // 前提条件表
        registry.put(PRECONDITION, new PatternSequencer(IDENTIFIED_ELEMENT__ID,
                ReferenceIndexSequencer.build(EFeaturePath.build(STPA_ANALYSIS__PRECONDITION), 1),
                Pattern.compile(".*-(\\d+)")));

        // アクシデント表
        registry.put(ACCIDENT,
                ReferenceIndexSequencer.build(EFeaturePath.build(STPA_ANALYSIS__ACCIDENT), 1));
        registry.put(HAZARD, ReferenceIndexSequencer.build(
                EFeaturePath.builder().add(STPA_ANALYSIS__ACCIDENT).add(ACCIDENT__HAZARD).create(),
                1));
        registry.put(SAFETY_CONSTRAINT,
                ReferenceIndexSequencer.build(EFeaturePath.builder().add(STPA_ANALYSIS__ACCIDENT)
                        .add(ACCIDENT__HAZARD).add(HAZARD__SAFETY_CONSTRAINT).create(), 1));

        // UCA表
        registry.put(CONTROL_ACTION, // CAには連番はないが、UCAから参照するために定義している
                ReferenceIndexSequencer.build(EFeaturePath.build(UCA_ANALYSIS__CONTROL_ACTION), 1));
        registry.put(UCA_GUIDE_WORD, new AttributeSequencer(UCA_GUIDE_WORD__UCA_ID_POSTFIX));
        registry.put(UNSAFE_CONTROL_ACTION, ReferenceIndexSequencer
                .build(EFeaturePath.build(UCA_ANALYSIS_RESULT__UNSAFE_CONTROL_ACTION), 1));

        // 対策表
        registry.put(COUNTERMEASURE,
                ReferenceIndexSequencer
                        .build(EFeaturePath.builder().add(STPA_ANALYSIS__HAZARD_CAUSAL_FACTOR)
                                .add(HAZARD_CAUSAL_FACTOR__COUNTERMEASURE).create(), 1));

        // HCF表
        // UCAに対する HCF自体の連番
        registry.put(HAZARD_CAUSAL_FACTOR, ReferenceIndexSequencer
                .builder(EFeaturePath.builder()
                        .addInverse(HAZARD_CAUSAL_FACTOR__UNSAFE_CONTROL_ACTION).create())
                .extraPaths(EFeaturePath.build(STPA_ANALYSIS__HAZARD_CAUSAL_FACTOR))
                .comparator(EComparators.eObjectPath()).offset(1).create());

        final IdReorganizer reorganizer = new IdReorganizer();

        // 前提条件表
        reorganizer.addFormat(registry.createFormat("{0}", PRECONDITION, IDENTIFIED_ELEMENT__ID));

        // アクシデント表
        reorganizer.addFormat(registry.createFormat("A{0}", ACCIDENT, IDENTIFIED_ELEMENT__ID));
        reorganizer.addFormat(registry.createFormat("H{0}", HAZARD, IDENTIFIED_ELEMENT__ID));
        reorganizer.addFormat(
                registry.createFormat("SC{0}", SAFETY_CONSTRAINT, IDENTIFIED_ELEMENT__ID));

        // UCA表
        reorganizer.addFormat(registry.createFormat("UCA{1}-{2}-{0}", UNSAFE_CONTROL_ACTION,
                IDENTIFIED_ELEMENT__ID,
                EFeaturePath.builder().add(UNSAFE_CONTROL_ACTION__CONTROL_ACTION)
                        .addInverse(UCA_ANALYSIS__CONTROL_ACTION).create(),
                EFeaturePath.builder().addInverse(UCA_ANALYSIS_RESULT__UNSAFE_CONTROL_ACTION)
                        .add(UCA_ANALYSIS_RESULT__GUIDE_WORD).add(UCA_GUIDE_WORD__UCA_ID_POSTFIX)
                        .create()));

        // 対策表
        reorganizer
                .addFormat(registry.createFormat("M{0}", COUNTERMEASURE, IDENTIFIED_ELEMENT__ID));

        // HCF表
        // UCAのIDを含む形式のため、多少非効率ですが、UCAと同様の依存関係を定義します
        reorganizer.addFormat(registry.createFormat("HCF{1}-{2}-{3}-{0}", HAZARD_CAUSAL_FACTOR,
                IDENTIFIED_ELEMENT__ID,
                EFeaturePath.builder().add(HAZARD_CAUSAL_FACTOR__UNSAFE_CONTROL_ACTION)
                        .add(UNSAFE_CONTROL_ACTION__CONTROL_ACTION)
                        .addInverse(UCA_ANALYSIS__CONTROL_ACTION).create(),
                EFeaturePath.builder().add(HAZARD_CAUSAL_FACTOR__UNSAFE_CONTROL_ACTION)
                        .addInverse(UCA_ANALYSIS_RESULT__UNSAFE_CONTROL_ACTION)
                        .add(UCA_ANALYSIS_RESULT__GUIDE_WORD).add(UCA_GUIDE_WORD__UCA_ID_POSTFIX)
                        .create(),
                EFeaturePath.builder().add(HAZARD_CAUSAL_FACTOR__UNSAFE_CONTROL_ACTION)
                        .add(IDENTIFIED_ELEMENT__ID).create()));

        return reorganizer;
    }

}
