package net.astah.stpa.stamp.notation.editor;

import com.google.common.collect.Iterables;

import net.astah.dd.dc.Bounds;
import net.astah.golf.draw.GGeometry;
import net.astah.golf.draw.GNode;
import net.astah.golf.draw.GRect;
import net.astah.golf.geom2D.Rectangle2d;
import net.astah.jomt.jview.HyperLinkIcon;
import net.astah.jomt.jview.HyperLinkIconImpl;
import net.astah.notation.editor.GraphicalStyles;
import net.astah.notation.editor.IDiagramViewElement;
import net.astah.notation.editor.ViewElements;
import net.astah.notation.editor.adjuster.MovingAdjuster;
import net.astah.stpa.stamp.notation.STAMPComponentShape;

/**
 * コンポーネント要素を表示するためのビューエレメントを実装する抽象クラスです。
 */
abstract public class STAMPComponentBaseViewElement
        extends STAMPShapeViewElementImpl<STAMPComponentShape> {

    /** 本体ノード */
    protected GRect bodyRect;

    /** ハイパーリンクアイコン */
    protected HyperLinkIcon hyperLinkIcon;

    /** ムービングアジャスターファクトリー */
    protected MovingAdjuster.Factory movingAdjusterFactory;

    public STAMPComponentBaseViewElement(STAMPComponentShape model) {
        super(model);
    }

    @Override
    public void refresh() {
        super.refresh();
        init();
        updateRimmedPick();
    }

    /**
     * 初期化処理を行います。
     */
    protected void init() {
        initBodyRect();
        initDrawSuggest();
        initHyperLinkIcon();
        initAdjusterFactory();
    }

    @Override
    public <T> T getAdapter(Class<T> adapter) {
        if (adapter.isInstance(hyperLinkIcon)) {
            return adapter.cast(hyperLinkIcon);
        } else if (adapter.isInstance(movingAdjusterFactory)) {
            return adapter.cast(movingAdjusterFactory);
        }
        return super.getAdapter(adapter);
    }

    /**
     * 本体ノードの初期化処理を行います。
     */
    protected void initBodyRect() {
        if (bodyRect == null) {
            int type = 0;
            type = type | RECT; // 2
            type = type | BODY; // 0
            bodyRect = new GRect();
            bodyRect.setUseIgnoreRimmedPick(true);
            bodyRect.setType(type);
            bodyRect.setFilled(true);
            bodyRect.setModel(getModel());
            bodyRect.addNodeSelectionListener(bodyRect);
            bodyRect.setSelectionType((byte) (GNode.KNOB | GNode.SHADOW));
            addNodeSelectionListener(bodyRect);
            addNode(bodyRect);
        }
    }

    /**
     * ドローサジェストの初期化処理を行います。
     */
    protected void initDrawSuggest() {
    }

    /**
     * ハイパーリンクアイコンの初期化処理を行います。
     */
    protected void initHyperLinkIcon() {
        if (hyperLinkIcon != null) {
            return;
        }
        hyperLinkIcon = new HyperLinkIconImpl(this, getModel());
        hyperLinkIcon.addIconNode();
    }

    /**
     * アジャスターファクトリーの初期化処理を行います。
     */
    protected void initAdjusterFactory() {
        movingAdjusterFactory = ViewElements.getAncestor(this, IDiagramViewElement.class)
                .getAdapter(MovingAdjuster.Factory.class);
    }

    @Override
    public void update() {
        super.update();
        updateBodyRect();
        updateHyperLinkIcon();
    }

    /**
     * 本体ノードの更新処理を行います。
     */
    protected void updateBodyRect() {
        bodyRect.setVisible(true);
        bodyRect.setPickable(true);
        bodyRect.setSelectable(true);
    }

    /**
     * ピック判定区分の更新処理を行います。
     */
    protected void updateRimmedPick() {
        bodyRect.setRimmedPick(Iterables.any(getOwnedElements(), getModel().eClass()::isInstance));
    }

    @Override
    protected void updateStroke() {
        GraphicalStyles.applyStrokeWidth(bodyRect, getModel(), 2);
    }

    @Override
    abstract protected void updateColor();

    /**
     * ハイパーリンクアイコンの更新処理を行います。
     */
    protected void updateHyperLinkIcon() {
        hyperLinkIcon.update();
    }

    /**
     * ハイパーリンクアイコンのレイアウト計算を行います。
     */
    protected void layoutHyperLinkIcon() {
        hyperLinkIcon.relocate(bodyRect);
    }

    @Override
    protected void layout(Bounds bounds) {
        layoutBodyRect(bounds);
        layoutHyperLinkIcon();
    }

    /**
     * 本体ノードのレイアウト計算を行います。
     * 
     * @param bounds
     *            バウンズ
     */
    protected void layoutBodyRect(Bounds bounds) {
        bodyRect.setRect(bounds.getX(), bounds.getY(), bounds.getWidth(), bounds.getHeight());
    }

    @Override
    protected GGeometry getBounds(Rectangle2d bounds) {
        return super.getBounds(bodyRect, bounds);
    }

    @Override
    public void setSelected() {
        bodyRect.setSelected(true);
    }

}
