package net.astah.stpa.stamp.notation.editor;

import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.command.UnexecutableCommand;
import org.eclipse.emf.ecore.EObject;

import lombok.NonNull;
import net.astah.dd.dc.Bounds;
import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.edit.provider.LayoutCalculator;
import net.astah.dd.di.util.DiagramElements;
import net.astah.golf.draw.GNode;
import net.astah.golf.geom2D.Pnt2d;
import net.astah.notation.Shape;
import net.astah.notation.editor.IComposableTransformer;
import net.astah.notation.editor.IShapeViewElement;
import net.astah.notation.editor.ITransformer;
import net.astah.notation.editor.ShapeTransformer;
import net.astah.stpa.stamp.ProcessValue;
import net.astah.stpa.stamp.ProcessVariable;

/**
 * ラベル要素の移動・リサイズを実装するクラスです。
 */
public class STAMPLabelTransformer extends ShapeTransformer implements IComposableTransformer {

    public STAMPLabelTransformer(@NonNull final IShapeViewElement target,
            @NonNull final Pnt2d startLocation) {
        super(target, startLocation);
    }

    public STAMPLabelTransformer(@NonNull final IShapeViewElement target,
            @NonNull final Pnt2d startLocation, int knob) {
        super(target, startLocation, knob);
    }

    @Override
    protected boolean isRestrictedBySize(Shape shape, Bounds bounds) {
        return true;
    }

    @Override
    public GNode createFeedbackNode(@NonNull final LayoutCalculator layoutCalculator) {
        return ProcessModelElements.isProcessModelElement((DiagramElement) getTarget().getModel())
                ? new GNode()
                : super.createFeedbackNode(layoutCalculator);
    }

    @Override
    public void move(@NonNull final Pnt2d endLocation,
            @NonNull final LayoutCalculator layoutCalculator, final GNode hoveredNode,
            @NonNull final Set<?> constraints) {
        final List<EObject> models = targets.stream()
                .map(v -> DiagramElements.getModelElement((DiagramElement) v.getModel()))
                .collect(Collectors.toList());
        if (models.stream()
                .anyMatch(m -> m instanceof ProcessVariable || m instanceof ProcessValue)) {
            command = move(models, endLocation, hoveredNode, layoutCalculator);
        } else {
            super.move(endLocation, layoutCalculator, hoveredNode, constraints);
        }
    }

    @Override
    protected Command move(final Shape shape, final Pnt2d endLocation, final GNode hoveredNode,
            final LayoutCalculator layoutCalculator) {
        final EObject selectedModel = DiagramElements.getModelElement(shape);
        if (selectedModel instanceof ProcessVariable || selectedModel instanceof ProcessValue) {
            final Shape destinationLabel = ((STAMPLabelViewElement) getTarget())
                    .getDestinationLabelToReorder(endLocation);
            return destinationLabel != null ? ProcessModelElements.createCommandToReorder(
                    Collections.singletonList(selectedModel), destinationLabel, endLocation.y)
                    : UnexecutableCommand.INSTANCE;
        } else {
            return super.move(shape, endLocation, hoveredNode, layoutCalculator);
        }
    }

    private Command move(final List<EObject> models, final Pnt2d endLocation,
            final GNode hoveredNode, final LayoutCalculator layoutCalculator) {
        final Shape destinationLabel = ((STAMPLabelViewElement) getTarget())
                .getDestinationLabelToReorder(endLocation);
        return destinationLabel != null
                ? ProcessModelElements.createCommandToReorder(models, destinationLabel,
                        endLocation.y)
                : UnexecutableCommand.INSTANCE;
    }

    @Override
    public ITransformer merge(ITransformer other) {
        if (!(other instanceof STAMPLabelTransformer)) {
            return null;
        }
        final EObject model = DiagramElements
                .getModelElement((DiagramElement) getTarget().getModel());
        if (!(model instanceof ProcessVariable) && !(model instanceof ProcessValue)) {
            return null;
        }
        other.getTargets().stream().filter(t -> !targets.contains(t))
                .map(it -> (IShapeViewElement) it).forEach(v -> targets.add(v));
        Collections.sort(targets);
        return this;
    }

}
