package net.astah.stpa.stamp.ui.control;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.eclipse.emf.common.command.CompoundCommand;
import org.eclipse.emf.common.util.AbstractTreeIterator;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.emf.transaction.TransactionalEditingDomain;

import com.google.common.collect.Iterators;
import com.google.common.collect.Lists;

import net.astah.dd.dc.DCFactory;
import net.astah.dd.di.Diagram;
import net.astah.dd.di.DiagramElement;
import net.astah.dd.di.Shape;
import net.astah.dd.di.edit.provider.LayoutCalculator;
import net.astah.dd.di.util.DiagramElements;
import net.astah.emf.edit.command.DestroyCommand;
import net.astah.notation.NotationFactory;
import net.astah.stpa.stamp.Component;
import net.astah.stpa.stamp.ControlStructure;
import net.astah.stpa.stamp.Link;
import net.astah.stpa.stamp.LinkEnd;
import net.astah.stpa.stamp.LinkableElement;
import net.astah.stpa.stamp.notation.STAMPEdge;
import net.astah.stpa.stamp.notation.STAMPNotationFactory;
import net.astah.stpa.stamp.notation.STAMPShape;
import net.astah.stpa.stamp.notation.util.STAMPDiagramElements;
import net.astah.ui.control.CreateEdgeCommand;
import net.astah.ui.control.CreateShapeCommand;

/**
 * 対象のコントロールストラクチャー図をコンポーネントの変更ベースに更新しするコマンドです。
 */
public class ModifyControlStructureDiagramElementsCommand extends CompoundCommand {

    TransactionalEditingDomain domain;
    Diagram diagram;
    List<LinkableElement> elements;
    List<Link> links;
    Point2D point;

    // 同期時に新規に作成した図要素のみ再レイアウトの対象とする
    private final List<DiagramElement> layoutTargets = new ArrayList<>();

    /**
     * コンストラクタ
     * 
     * @param domain
     *            domain
     * @param controlStructure
     *            更新対象のコントロールストラクチャー図
     */
    public ModifyControlStructureDiagramElementsCommand(TransactionalEditingDomain domain,
            ControlStructure controlStructure) {
        this(domain,
                Lists.newArrayList(Iterators.filter(
                        EcoreUtil.getAllContents(controlStructure, false), LinkableElement.class)),
                controlStructure);
    }

    /**
     * コンストラクタ
     * 
     * @param domain
     *            domain
     * @param elements
     *            更新対象のコンポーネント群
     * @param controlStructure
     *            更新対象のコントロールストラクチャー図
     */
    public ModifyControlStructureDiagramElementsCommand(TransactionalEditingDomain domain,
            List<LinkableElement> elements, ControlStructure controlStructure) {
        this(domain, elements, controlStructure, null);
    }

    /**
     * コンストラクタ
     * 
     * @param domain
     *            domain
     * @param elements
     *            更新対象のコンポーネント群
     * @param controlStructure
     *            更新対象のコントロールストラクチャー図
     * @param point
     *            図要素作成の初期位置
     */
    public ModifyControlStructureDiagramElementsCommand(TransactionalEditingDomain domain,
            List<LinkableElement> elements, ControlStructure controlStructure, Point2D point) {
        this.domain = domain;
        diagram = STAMPDiagramElements.getDiagramElements(controlStructure, null, Diagram.class)
                .stream().findFirst().orElse(null);

        this.elements = elements;
        links = controlStructure.getLinks();
        this.point = point;
    }

    @Override
    protected boolean prepare() {
        return diagram != null;
    }

    @Override
    public void execute() {
        final LayoutCalculator layoutCalculator = new LayoutCalculator();

        // コマンドの作成を遅延させ、始点・終点→コンポーネント→リンク→レイアウトの順番にコマンドが作成、実行する
        createLinkEnds(layoutCalculator);
        deleteComponentShapes();
        createComponentShapes(layoutCalculator);
        createEdges();
        appendAndExecute(new ControlStructureDiagramLayoutCommand(domain, diagram, layoutTargets));
        return;
    }

    /**
     * コンポーネントの状況に応じて、コンポーネントの図要素を削除します。
     */
    private void deleteComponentShapes() {
        Set<STAMPShape> componentShapes = new HashSet<>();
        for (LinkableElement element : elements) {
            if (!(element instanceof Component)) {
                continue;
            }
            Component component = (Component) element;
            if (!isAnalysisTarget(component)) {
                continue;
            }
            componentShapes.addAll(
                    STAMPDiagramElements.getDiagramElements(component, diagram, STAMPShape.class));
        }

        TreeIterator<DiagramElement> componentIterator = new AbstractTreeIterator<DiagramElement>(
                diagram, false) {
            private static final long serialVersionUID = 1L;

            @Override
            public Iterator<DiagramElement> getChildren(final Object object) {
                return Iterators.filter(((DiagramElement) object).getOwnedElements().iterator(),
                        m -> {
                            return m instanceof STAMPShape && DiagramElements
                                    .getModelElement(m) instanceof Component;
                        });
            }
        };

        List<DiagramElement> deleteShapes = Lists.newArrayList(componentIterator);
        deleteShapes.removeAll(componentShapes);

        if (!deleteShapes.isEmpty()) {
            appendAndExecute(new DestroyCommand(domain, deleteShapes));
        }
    }

    /**
     * コンポーネントの状態に応じて、コンポーネントの図要素を作成します。
     */
    private void createComponentShapes(final LayoutCalculator layoutCalculator) {
        for (LinkableElement element : elements) {
            if (!(element instanceof Component)) {
                continue;
            }
            Component component = (Component) element;
            if (!isAnalysisTarget(component)) {
                continue;
            }

            if (!existDiagramElement(component, STAMPShape.class)) {
                CreateShapeCommand cmd = new CreateShapeCommand();
                Shape parent = STAMPDiagramElements
                        .getDiagramElements(component.eContainer(), diagram, STAMPShape.class)
                        .stream().map(m -> (Shape) m).findFirst().orElse(diagram);

                cmd.setOwningElement(parent);
                final STAMPShape shape = STAMPNotationFactory.eINSTANCE.createSTAMPComponentShape();
                shape.getModelElements().add(component);
                cmd.setShape(shape);
                cmd.setLocation(DCFactory.eINSTANCE.createPoint(0, 0));
                appendAndExecute(cmd.createTransactionalCommand(domain));
                layoutTargets.add(shape);
            }
        }
    }

    /**
     * 始点・終点の図要素を作成します。
     */
    private void createLinkEnds(final LayoutCalculator layoutCalculator) {
        for (LinkableElement element : elements) {
            if (!(element instanceof LinkEnd)) {
                continue;
            }
            LinkEnd linkEnd = (LinkEnd) element;

            if (!existDiagramElement(linkEnd, STAMPShape.class)) {
                CreateShapeCommand cmd = new CreateShapeCommand();
                Shape parent = STAMPDiagramElements
                        .getDiagramElements(linkEnd.getOwner(), diagram, STAMPShape.class)
                        .stream().map(m -> (Shape) m).findFirst().orElse(diagram);

                cmd.setOwningElement(parent);
                final STAMPShape shape = STAMPNotationFactory.eINSTANCE.createSTAMPShape();
                shape.getModelElements().add(linkEnd);
                cmd.setShape(shape);
                cmd.setLocation(DCFactory.eINSTANCE.createPoint(0, 0));
                appendAndExecute(cmd.createTransactionalCommand(domain));
                layoutTargets.add(shape);
            }
        }
    }

    /**
     * コントロールストラクチャー図内のリンクの図要素を作成します。 LinkSource 、 LinkTarget の図要素が存在しない場合は作成しません。
     */
    private void createEdges() {
        for (Link link : links) {

            STAMPShape sourceShape = STAMPDiagramElements
                    .getDiagramElements(link.getSource(), diagram, STAMPShape.class).stream()
                    .findFirst().orElse(null);
            STAMPShape targetShape = STAMPDiagramElements
                    .getDiagramElements(link.getTarget(), diagram, STAMPShape.class).stream()
                    .findFirst().orElse(null);

            if (targetShape == null || sourceShape == null) {
                continue;
            }

            if (!existDiagramElement(link, STAMPEdge.class)) {
                final CreateEdgeCommand cmd = new CreateEdgeCommand();
                final STAMPEdge edge = STAMPNotationFactory.eINSTANCE.createSTAMPEdge();
                edge.getModelElements().add(link);
                cmd.setEdge(edge);
                cmd.setSource(sourceShape);
                cmd.setTarget(targetShape);
                cmd.setLine(NotationFactory.eINSTANCE.createRightAngleLine());
                appendAndExecute(cmd.createTransactionalCommand(domain));
                layoutTargets.add(edge);
            }
        }
    }

    private <T extends DiagramElement> boolean existDiagramElement(final EObject model,
            final Class<T> targetClass) {
        final List<T> elements = STAMPDiagramElements.getDiagramElements(model, diagram,
                targetClass);
        return !elements.isEmpty();
    }

    /**
     * 入れ子を考慮してコンポーネントが分析対象かどうかを返します。
     * 
     * @param component
     *            コンポーネント
     * @return boolean コンポーネントが分析対象かどうか
     */
    private boolean isAnalysisTarget(Component component) {
        boolean containerResult = component.eContainer() instanceof Component
                ? isAnalysisTarget((Component) component.eContainer())
                : true;
        return component.isAnalysisTarget() && containerResult;
    }

}
