package net.astah.stpa.stamp.ui.menu.handler;

import static net.astah.stpa.stamp.STAMPPackage.Literals.ACCIDENT__HAZARD;
import static net.astah.table.TablePackage.Literals.ELEMENT__OBJECT;
import static org.eclipse.emf.edit.domain.AdapterFactoryEditingDomain.getEditingDomainFor;

import java.awt.event.ActionEvent;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JFrame;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.command.CompoundCommand;
import org.eclipse.emf.common.command.UnexecutableCommand;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.edit.command.AddCommand;
import org.eclipse.emf.edit.command.SetCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.emf.transaction.util.TransactionUtil;

import net.astah.emf.edit.command.DestroyCommand;
import net.astah.emf.edit.job.CommandJob;
import net.astah.golf.widget.IListSelection;
import net.astah.golf.widget.ISelection;
import net.astah.golf.widget.ITableSelection;
import net.astah.golf.widget.swing.SwingFrame;
import net.astah.jomt.jsystem.JomtGlobal;
import net.astah.jomt.jview.Editor;
import net.astah.jomt.jview.EditorPeer;
import net.astah.stpa.stamp.Accident;
import net.astah.stpa.stamp.Hazard;
import net.astah.stpa.stamp.IdentifiedElement;
import net.astah.stpa.stamp.STPAAnalysis;
import net.astah.stpa.stamp.ui.Messages;
import net.astah.stpa.stamp.ui.view.SelectElementDialog;
import net.astah.stpa.stamp.ui.view.swing.AccidentHazardSafetyConstraintTableEditorPeer;
import net.astah.table.Element;
import net.astah.table.Tuple;
import net.astah.ui.menu.annotation.CanExecute;
import net.astah.ui.menu.annotation.Execute;
import net.astah.ui.menu.annotation.Handler;
import net.astah.ui.menu.annotation.MenuCommand;
import net.astah.ui.menu.annotation.MenuItem;
import net.astah.ui.menu.annotation.MenuItem.ItemKind;

/**
 * アクシデント表で右クリックしたハザードのセルに設定するハザードを、ハザード選択ダイアログで指定します。
 * 
 * <p>
 * 同じアクシデントに同じハザードを複数指定することはできません。
 */
@Handler
@MenuItem(commands = @MenuCommand(id = "stpa.SelectHazardCommand",
        category = "edit.table"),
        label = "%StampTable_selectHazard_label",
        tooltip = "%StampTable_selectHazard_tooltip",
        type = ItemKind.PUSH)
public class SelectHazardHandler {

    private final int HAZARD_ID_COLUMN = 2;
    private final int HAZARD_COLUMN = 3;

    private Command command = UnexecutableCommand.INSTANCE;

    /**
     * ハザード選択ダイアログで指定したハザードをアクシデントに設定するコマンドを実行します。
     * 
     * @param selection
     *            選択要素
     * @param params
     *            コマンドのパラメータ
     * @param e
     *            アクションイベント
     */
    @Execute
    public void execute(final ISelection selection, final List<String> params,
            final ActionEvent e) {
        if (selection instanceof ITableSelection) {
            final ITableSelection tableSelection = (ITableSelection) selection;
            final IListSelection rowSelection = tableSelection.getRowSelection();
            command = selectHazard((Tuple) rowSelection.getFirstElement());
            final CommandJob job = new CommandJob(getDomain(), command);
            job.schedule();
        }
    }

    private TransactionalEditingDomain getDomain() {
        final Editor editor = JomtGlobal.viewManager.getCurrentEditor();
        return TransactionUtil.getEditingDomain(editor.getEditModel());
    }

    /**
     * アクシデント表の右クリックしたセルでハザードを選択できるかどうか確認します。
     * 
     * @param selection
     *            選択要素
     * @return true アクシデント表にハザードが存在する、かつ、ハザードのセルを右クリックした場合 / false 上記以外
     */
    @CanExecute
    public boolean canExecute(final ISelection selection) {
        return selection instanceof ITableSelection
                && isTarget((ITableSelection) selection);
    }

    private boolean isTarget(final ITableSelection selection) {
        final Object element = selection.getRowSelection().getFirstElement();
        if (element instanceof Tuple) {
            return existHazards()
                    && isHazard(selection.getColumnSelection());
        }
        return false;
    }

    private boolean existHazards() {
        final AccidentHazardSafetyConstraintTableEditorPeer editorPeer = getEditorPeer();
        final STPAAnalysis scope = editorPeer.getScope();
        return !scope.getHazards().isEmpty();
    }

    private boolean isHazard(final IListSelection columnSelection) {
        final int[] indices = columnSelection.getIndices();
        if (indices.length > 0) {
            return indices[0] == HAZARD_ID_COLUMN
                    || indices[0] == HAZARD_COLUMN;
        }
        return false;
    }

    private Command selectHazard(final Tuple rowTuple) {
        final AccidentHazardSafetyConstraintTableEditorPeer editorPeer = getEditorPeer();
        final Accident accident = (Accident) rowTuple.getElements().get(editorPeer.ACCIDENT_INDEX)
                .getObject();
        final JFrame parent = (JFrame) ((SwingFrame) JomtGlobal.viewManager.getManagementView()
                .getComponent()).getComponent();

        final List<IdentifiedElement> list = new ArrayList<>(editorPeer.getScope().getHazards());
        final SelectElementDialog dialog = new SelectElementDialog(parent, list,
                (IdentifiedElement) rowTuple.getElements().get(editorPeer.HAZARD_INDEX)
                        .getObject());
        dialog.setTitle(accident.getId() + " "
                + Messages.accident_hazard_safety_constraint_table_select_hazard_label);
        dialog.pack();
        dialog.setVisible(true);

        if (dialog.getResultState() == SelectElementDialog.OK) {
            return createSetHazardCommand(rowTuple, (Hazard) dialog.getSelectedElement(),
                    editorPeer);
        }
        return command;
    }

    private Command createSetHazardCommand(final Tuple rowTuple, final Hazard hazard,
            final AccidentHazardSafetyConstraintTableEditorPeer editorPeer) {
        final STPAAnalysis scope = editorPeer.getScope();
        final EditingDomain domain = getEditingDomainFor(scope);
        final EList<Element> elements = rowTuple.getElements();
        final EObject accident = elements.get(editorPeer.ACCIDENT_INDEX).getObject();
        final Hazard oldHazard = (Hazard) elements.get(editorPeer.HAZARD_INDEX).getObject();
        if (oldHazard == hazard) {
            return UnexecutableCommand.INSTANCE;
        }

        final CompoundCommand command = new CompoundCommand();
        if (oldHazard != null) {
            command.append(DestroyCommand.create(domain,
                    new DestroyCommand.Reference(accident, ACCIDENT__HAZARD, oldHazard)));
        }
        command.append(AddCommand.create(domain, accident, ACCIDENT__HAZARD, hazard));
        command.append(SetCommand.create(getDomain(), elements.get(editorPeer.HAZARD_INDEX),
                ELEMENT__OBJECT, hazard));

        return command;
    }

    private AccidentHazardSafetyConstraintTableEditorPeer getEditorPeer() {
        final Editor editor = JomtGlobal.viewManager.getCurrentEditor();
        final EditorPeer editorPeer = editor.getEditorPeer();
        return (AccidentHazardSafetyConstraintTableEditorPeer) editorPeer;
    }

}
