package net.astah.stpa.stamp.ui.menu.handler;

import static net.astah.stpa.stamp.STAMPPackage.Literals.HAZARD__SAFETY_CONSTRAINT;
import static net.astah.table.TablePackage.Literals.ELEMENT__OBJECT;
import static org.eclipse.emf.edit.domain.AdapterFactoryEditingDomain.getEditingDomainFor;

import java.awt.event.ActionEvent;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JFrame;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.command.CompoundCommand;
import org.eclipse.emf.common.command.UnexecutableCommand;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.edit.command.AddCommand;
import org.eclipse.emf.edit.command.SetCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.emf.transaction.util.TransactionUtil;

import net.astah.emf.edit.command.DestroyCommand;
import net.astah.emf.edit.job.CommandJob;
import net.astah.golf.widget.IListSelection;
import net.astah.golf.widget.ISelection;
import net.astah.golf.widget.ITableSelection;
import net.astah.golf.widget.swing.SwingFrame;
import net.astah.jomt.jsystem.JomtGlobal;
import net.astah.jomt.jview.Editor;
import net.astah.jomt.jview.EditorPeer;
import net.astah.stpa.stamp.Hazard;
import net.astah.stpa.stamp.IdentifiedElement;
import net.astah.stpa.stamp.STPAAnalysis;
import net.astah.stpa.stamp.SafetyConstraint;
import net.astah.stpa.stamp.ui.Messages;
import net.astah.stpa.stamp.ui.view.SelectElementDialog;
import net.astah.stpa.stamp.ui.view.swing.AccidentHazardSafetyConstraintTableEditorPeer;
import net.astah.table.Element;
import net.astah.table.Tuple;
import net.astah.ui.menu.annotation.CanExecute;
import net.astah.ui.menu.annotation.Execute;
import net.astah.ui.menu.annotation.Handler;
import net.astah.ui.menu.annotation.MenuCommand;
import net.astah.ui.menu.annotation.MenuItem;
import net.astah.ui.menu.annotation.MenuItem.ItemKind;

/**
 * アクシデント表で右クリックした安全制約のセルに設定する安全制約を、安全制約選択ダイアログで指定します。
 * <ul>
 * <li>同じ行にハザードがない場合、安全制約は選択できません。
 * <li>同じハザードに同じ安全制約を複数指定することはできません。
 * </ul>
 */
@Handler
@MenuItem(commands = @MenuCommand(id = "stpa.SelectSafetyConstraintCommand",
        category = "edit.table"),
        label = "%StampTable_selectSafetyConstraint_label",
        tooltip = "%StampTable_selectSafetyConstraint_tooltip",
        type = ItemKind.PUSH)
public class SelectSafetyConstraintHandler {

    private final int SAFETY_CONSTRAINT_ID_COLUMN = 4;

    private Command command = UnexecutableCommand.INSTANCE;

    /**
     * 安全制約選択ダイアログで指定した安全制約をハザードに設定するコマンドを実行します。
     * 
     * @param selection
     *            選択要素
     * @param params
     *            コマンドのパラメータ
     * @param e
     *            アクションイベント
     */
    @Execute
    public void execute(final ISelection selection, final List<String> params,
            final ActionEvent e) {
        if (selection instanceof ITableSelection) {
            final ITableSelection tableSelection = (ITableSelection) selection;
            final IListSelection rowSelection = tableSelection.getRowSelection();
            command = selectSafetyConstraint((Tuple) rowSelection.getFirstElement());
            final CommandJob job = new CommandJob(getDomain(), command);
            job.schedule();
        }
    }

    private TransactionalEditingDomain getDomain() {
        final Editor editor = JomtGlobal.viewManager.getCurrentEditor();
        return TransactionUtil.getEditingDomain(editor.getEditModel());
    }

    /**
     * アクシデント表の右クリックしたセルで安全制約を選択できるかどうか確認します。
     * 
     * @param selection
     *            選択要素
     * @return true 右クリックしたセルの行にハザードが存在する、かつ、アクシデント表に安全制約が存在する かつ、安全制約のセルを右クリックした場合
     *         / false 上記以外
     */
    @CanExecute
    public boolean canExecute(final ISelection selection) {
        return selection instanceof ITableSelection
                && isTarget((ITableSelection) selection);
    }

    private boolean isTarget(final ITableSelection selection) {
        final Object element = selection.getRowSelection().getFirstElement();
        if (element instanceof Tuple) {
            final AccidentHazardSafetyConstraintTableEditorPeer editorPeer = getEditorPeer();
            return (hasHazard((Tuple) element, editorPeer)
                    && existSafetyConstraints(editorPeer)
                    && isSafetyConstraint(selection.getColumnSelection()));
        }
        return false;
    }

    private boolean hasHazard(final Tuple tuple,
            final AccidentHazardSafetyConstraintTableEditorPeer editorPeer) {
        return tuple.getElements().get(editorPeer.HAZARD_INDEX).getObject() != null;
    }

    private boolean existSafetyConstraints(
            final AccidentHazardSafetyConstraintTableEditorPeer editorPeer) {
        final STPAAnalysis scope = editorPeer.getScope();
        return !scope.getSafetyConstraints().isEmpty();
    }

    private boolean isSafetyConstraint(final IListSelection columnSelection) {
        final int[] indices = columnSelection.getIndices();
        if (indices.length > 0) {
            return indices[0] >= SAFETY_CONSTRAINT_ID_COLUMN;
        }
        return false;
    }

    private Command selectSafetyConstraint(final Tuple rowTuple) {
        final AccidentHazardSafetyConstraintTableEditorPeer editorPeer = getEditorPeer();
        final Hazard hazard = (Hazard) rowTuple.getElements().get(editorPeer.HAZARD_INDEX)
                .getObject();
        final JFrame parent = (JFrame) ((SwingFrame) JomtGlobal.viewManager.getManagementView()
                .getComponent()).getComponent();

        final List<IdentifiedElement> list = new ArrayList<>(
                editorPeer.getScope().getSafetyConstraints());
        final SelectElementDialog dialog = new SelectElementDialog(parent, list,
                (IdentifiedElement) rowTuple.getElements().get(editorPeer.SAFETY_CONSTRAINT_INDEX)
                        .getObject());
        dialog.setTitle(hazard.getId() + " "
                + Messages.accident_hazard_safety_constraint_table_select_safety_constraint_label);
        dialog.pack();
        dialog.setVisible(true);

        if (dialog.getResultState() == SelectElementDialog.OK) {
            return createSetSafetyConstraintCommand(rowTuple,
                    (SafetyConstraint) dialog.getSelectedElement(), editorPeer);
        }
        return command;
    }

    private Command createSetSafetyConstraintCommand(final Tuple rowTuple,
            final SafetyConstraint safetyConstraint,
            final AccidentHazardSafetyConstraintTableEditorPeer editorPeer) {
        final STPAAnalysis scope = editorPeer.getScope();
        final EditingDomain domain = getEditingDomainFor(scope);
        final EList<Element> elements = rowTuple.getElements();
        final EObject hazard = elements.get(editorPeer.HAZARD_INDEX).getObject();
        final SafetyConstraint oldSafetyConstraint = (SafetyConstraint) elements
                .get(editorPeer.SAFETY_CONSTRAINT_INDEX).getObject();

        final CompoundCommand command = new CompoundCommand();

        if (oldSafetyConstraint != null) {
            if (oldSafetyConstraint != safetyConstraint
                    && oldSafetyConstraint.getHazards().size() == 1) {
                command.append(DestroyCommand.create(domain, oldSafetyConstraint));
                command.append(AddCommand.create(domain, hazard, HAZARD__SAFETY_CONSTRAINT,
                        safetyConstraint));
            }
        } else {
            command.append(AddCommand.create(domain, hazard,
                    HAZARD__SAFETY_CONSTRAINT, safetyConstraint));
        }

        command.append(SetCommand.create(getDomain(),
                elements.get(editorPeer.SAFETY_CONSTRAINT_INDEX),
                ELEMENT__OBJECT, safetyConstraint));

        return command.unwrap();
    }

    private AccidentHazardSafetyConstraintTableEditorPeer getEditorPeer() {
        final Editor editor = JomtGlobal.viewManager.getCurrentEditor();
        final EditorPeer editorPeer = editor.getEditorPeer();
        return (AccidentHazardSafetyConstraintTableEditorPeer) editorPeer;
    }

}
