package net.astah.stpa.stamp.ui.menu.handler;

import java.awt.event.ActionEvent;
import java.util.List;

import javax.swing.JMenuItem;

import org.eclipse.emf.ecore.EObject;

import net.astah.dd.dg.Line;
import net.astah.dd.dg.MarkedElement;
import net.astah.dd.dg.Marker;
import net.astah.dd.dg.Polyline;
import net.astah.dd.dg.util.DGStyleResolver;
import net.astah.golf.control.GolfAction;
import net.astah.golf.widget.ISelection;
import net.astah.golf.widget.swing.Selections;
import net.astah.jmodel.StyleSupport;
import net.astah.notation.resource.NotationResources;
import net.astah.ui.menu.annotation.AboutToShow;
import net.astah.ui.menu.annotation.CanExecute;
import net.astah.ui.menu.annotation.Execute;
import net.astah.ui.menu.annotation.Handler;
import net.astah.ui.menu.annotation.MenuCommand;
import net.astah.ui.menu.annotation.MenuItem;
import net.astah.ui.menu.annotation.MenuItem.ItemKind;

/**
 * 図要素に矢印を設定します。
 * 
 * <p>
 * 対象とする要素 : 直線、フリーハンド。
 * <p>
 * 矢印の種類 : 矢印なし、始点矢印、終点矢印、両端矢印。
 */
@Handler(commands = "stpa.SetLineArrowCommand")
@MenuItem(commands = @MenuCommand(id = "stpa.SetLineArrowCommand",
        category = "edit.style"),
        path = "LineArrow",
        label = "%StampControlStructureDiagram_lineArrow_label",
        tooltip = "%StampControlStructureDiagram_lineArrow_tooltip",
        type = ItemKind.MENU)
@MenuItem(commands = @MenuCommand(id = "stpa.SetLineArrowCommand",
        category = "edit.style",
        parameters = StyleSupport.STYLE_LINE_ARROW_TYPE_NONE),
        path = "LineArrow." + StyleSupport.STYLE_LINE_ARROW_TYPE_NONE,
        iconURI = "platform:/plugin/net.astah.core.app/net/astah/jomt/resource/images/lineWithNoArrow.png",
        type = ItemKind.CHECK)
@MenuItem(commands = @MenuCommand(id = "stpa.SetLineArrowCommand",
        category = "edit.style",
        parameters = StyleSupport.STYLE_LINE_ARROW_TYPE_START),
        path = "LineArrow." + StyleSupport.STYLE_LINE_ARROW_TYPE_START,
        iconURI = "platform:/plugin/net.astah.core.app/net/astah/jomt/resource/images/lineWithLArrow.png",
        type = ItemKind.CHECK)
@MenuItem(commands = @MenuCommand(id = "stpa.SetLineArrowCommand",
        category = "edit.style",
        parameters = StyleSupport.STYLE_LINE_ARROW_TYPE_END),
        path = "LineArrow." + StyleSupport.STYLE_LINE_ARROW_TYPE_END,
        iconURI = "platform:/plugin/net.astah.core.app/net/astah/jomt/resource/images/lineWithRArrow.png",
        type = ItemKind.CHECK)
@MenuItem(commands = @MenuCommand(id = "stpa.SetLineArrowCommand",
        category = "edit.style",
        parameters = StyleSupport.STYLE_LINE_ARROW_TYPE_BOTH),
        path = "LineArrow." + StyleSupport.STYLE_LINE_ARROW_TYPE_BOTH,
        iconURI = "platform:/plugin/net.astah.core.app/net/astah/jomt/resource/images/lineWithArrows.png",
        type = ItemKind.CHECK)
public class SetLineArrowHandler {

    /**
     * 図要素に矢印を設定するコマンドを実行します。
     * 
     * @param selection
     *            選択要素
     * @param params
     *            コマンドのパラメータ
     * @param e
     *            アクションイベント
     */
    @Execute
    public void execute(final ISelection selection, final List<String> params,
            final ActionEvent e) {
        final GolfAction action = new GolfAction(
                "net.astah.jomt.jcontrol.SetLineArrow%" + params.get(0));
        action.actionPerformed(e);
    }

    /**
     * 図要素に矢印を設定するコマンドが実行可能かどうか確認します。
     * 
     * @param selection
     *            選択要素
     * @return true コマンドが実行可能 / false コマンドが実行不可
     */
    @CanExecute
    public boolean canExecute(final ISelection selection) {
        // 選択要素が直線、フリーハンドのみの場合、true
        return !Selections.isEmpty(selection)
                && Selections.stream(selection).allMatch(object -> isTarget(object));
    }

    private boolean isTarget(final Object object) {
        return object instanceof Line || isFreehand(object);
    }

    private boolean isFreehand(final Object object) {
        if (object instanceof Polyline) {
            final Polyline polyline = (Polyline) object;
            final double strokeOpacity = DGStyleResolver.getInstance(polyline)
                    .resolveStrokeOpacity(polyline);
            return strokeOpacity >= 1;
        }
        return false;
    }

    /**
     * メニュー表示の直前に、図要素の矢印の種類に該当するメニューをチェックします。
     * 
     * @param selection
     *            選択要素
     * @param menuItem
     *            表示するメニューアイテム
     * @param params
     *            コマンドに指定されているパラメータ
     */
    @AboutToShow
    public void aboutToShow(final ISelection selection, final JMenuItem menuItem,
            final List<String> params) {
        if (params.isEmpty()) {
            return;
        }
        if (Selections.size(selection) > 1) {
            return;
        }
        final boolean isSameLineArrow = isSameLineArrow(Selections.getFirstElement(selection),
                params.get(0));
        menuItem.setSelected(isSameLineArrow);
    }

    private boolean isSameLineArrow(final Object object, final String lineArrow) {
        if (object instanceof MarkedElement) {
            final MarkedElement element = (MarkedElement) object;
            final Marker start = element.getStartMarker();
            final Marker end = element.getEndMarker();
            final Marker startMarker = NotationResources.getLibraryMarker((EObject) object,
                    "Arrowhead-Start");
            final Marker endMarker = NotationResources.getLibraryMarker((EObject) object,
                    "Arrowhead-End");
            switch (lineArrow) {
            case StyleSupport.STYLE_LINE_ARROW_TYPE_NONE:
                return start == null && end == null;
            case StyleSupport.STYLE_LINE_ARROW_TYPE_END:
                return start == null && endMarker.equals(end);
            case StyleSupport.STYLE_LINE_ARROW_TYPE_START:
                return startMarker.equals(start) && end == null;
            case StyleSupport.STYLE_LINE_ARROW_TYPE_BOTH:
                return startMarker.equals(start) && endMarker.equals(end);
            }
        }
        return false;
    }

}