package net.astah.stpa.stamp.ui.menu.handler;

import java.awt.event.ActionEvent;
import java.util.Collections;
import java.util.List;

import javax.swing.JMenuItem;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.command.CompoundCommand;
import org.eclipse.emf.common.command.UnexecutableCommand;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.edit.command.AddCommand;
import org.eclipse.emf.edit.command.CommandParameter;
import org.eclipse.emf.edit.command.CreateChildCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.emf.transaction.util.TransactionUtil;

import net.astah.dd.di.util.DiagramElements;
import net.astah.emf.edit.command.DestroyCommand;
import net.astah.emf.edit.job.CommandJob;
import net.astah.golf.widget.ISelection;
import net.astah.golf.widget.swing.Selections;
import net.astah.stpa.stamp.Component;
import net.astah.stpa.stamp.ProcessModel;
import net.astah.stpa.stamp.STAMPFactory;
import net.astah.stpa.stamp.STAMPPackage;
import net.astah.stpa.stamp.notation.STAMPCompartment;
import net.astah.stpa.stamp.notation.STAMPComponentShape;
import net.astah.stpa.stamp.notation.STAMPNotationPackage;
import net.astah.stpa.stamp.notation.edit.command.SynchronizeProcessModelCommand;
import net.astah.ui.menu.annotation.AboutToShow;
import net.astah.ui.menu.annotation.CanExecute;
import net.astah.ui.menu.annotation.Execute;
import net.astah.ui.menu.annotation.Handler;
import net.astah.ui.menu.annotation.MenuCommand;
import net.astah.ui.menu.annotation.MenuItem;
import net.astah.ui.menu.annotation.MenuItem.ItemKind;

/**
 * プロセスモデル区画の表示非表示を切り替えます。
 */
@Handler
@MenuItem(label = "%StampControlStructureDiagram_showProcessModelCompartment_label",
        tooltip = "%StampControlStructureDiagram_showProcessModelCompartment_tooltip",
        type = ItemKind.CHECK)
@MenuCommand(id = "stpa.ShowProcessModelCompartmentCommand",
        category = "edit.show_diagram_element")
public class ShowProcessModelCompartmentHandler {
    private Command command = UnexecutableCommand.INSTANCE;

    /**
     * プロセスモデル区画の表示非表示を切り替えるコマンドを実行します。
     * 
     * @param selection
     *            選択要素
     * @param params
     *            コマンドのパラメータ
     * @param e
     *            アクションイベント
     */
    @Execute
    public void execute(final ISelection selection, final List<String> params,
            final ActionEvent e) {
        final STAMPComponentShape shape = (STAMPComponentShape) Selections
                .getFirstElement(selection);
        final Component component = DiagramElements.getModelElement(shape, Component.class);
        final TransactionalEditingDomain domain = TransactionUtil.getEditingDomain(component);
        final CommandJob job = new CommandJob(domain, command);
        job.schedule();
    }

    private Command getCommand(final ISelection selection) {
        if (Selections.size(selection) != 1) {
            return UnexecutableCommand.INSTANCE;
        }
        final Object element = Selections.getFirstElement(selection);
        if (!(element instanceof STAMPComponentShape)) {
            return UnexecutableCommand.INSTANCE;
        }
        final STAMPComponentShape shape = (STAMPComponentShape) element;
        if (shape.isOmitted()) {
            return UnexecutableCommand.INSTANCE;
        }
        final Component component = DiagramElements.getModelElement(shape, Component.class);
        final TransactionalEditingDomain domain = TransactionUtil.getEditingDomain(component);
        final ProcessModel processModel = component.getProcessModel();
        final STAMPCompartment compartment = Compartments
                .getProcessModelCompartment(shape.getCompartments(), processModel);
        return compartment == null
                ? createCreateCompartmentCommand(domain, component, processModel, shape)
                : createDeleteCompartmentCommand(domain, compartment);
    }

    private Command createCreateCompartmentCommand(
            final EditingDomain domain, final Component component,
            final ProcessModel existingProcessModel, final STAMPComponentShape shape) {
        final CompoundCommand command = new CompoundCommand();
        if (existingProcessModel == null) {
            final ProcessModel processModel = STAMPFactory.eINSTANCE.createProcessModel();
            command.append(CreateChildCommand.create(
                    domain, component, new CommandParameter(component,
                            STAMPPackage.Literals.COMPONENT__PROCESS_MODEL, processModel),
                    Collections.singleton(component)));
            final STAMPCompartment compartment = Compartments.createCompartment(processModel);
            command.append(AddCommand.create(domain, shape,
                    STAMPNotationPackage.Literals.STAMP_COMPARTMENTABLE_SHAPE__COMPARTMENT,
                    compartment));
        } else {
            final STAMPCompartment compartment = Compartments
                    .createCompartment(existingProcessModel);
            command.append(AddCommand.create(domain, shape,
                    STAMPNotationPackage.Literals.STAMP_COMPARTMENTABLE_SHAPE__COMPARTMENT,
                    compartment));
            command.appendIfCanExecute(new SynchronizeProcessModelCommand(domain, compartment));
        }
        return command;
    }

    private Command createDeleteCompartmentCommand(
            final EditingDomain domain, final STAMPCompartment compartment) {
        return DestroyCommand.create(domain, compartment);
    }

    /**
     * プロセスモデル区画の表示非表示を切り替えるコマンドが実行可能かどうか確認します。
     * 
     * @param selection
     *            選択要素
     * @return true コマンドが実行可能 / false コマンドが実行不可
     */
    @CanExecute
    public boolean canExecute(final ISelection selection) {
        command.dispose();
        command = getCommand(selection);
        return command.canExecute();
    }

    /**
     * メニュー表示の直前に、プロセスモデル区画の表示非表示に該当するメニューアイテムをチェックします。
     * 
     * @param selection
     *            選択要素
     * @param menuItem
     *            表示するメニューアイテム
     * @param params
     *            コマンドに指定されているパラメータ
     */
    @AboutToShow
    public void aboutToShow(final ISelection selection, final JMenuItem menuItem,
            final List<String> params) {
        final STAMPComponentShape shape = (STAMPComponentShape) Selections
                .getFirstElement(selection);
        final Component component = DiagramElements.getModelElement(shape, Component.class);
        final ProcessModel processModel = component.getProcessModel();
        menuItem.setSelected(hasProcessModel(shape.getCompartments(), processModel));
    }

    /**
     * 指定されたプロセスモデルに対応する区画が存在するかどうかを調べます。
     * 
     * @param compartments
     *            コンポーネントが持つ全ての区画
     * @param processModel
     *            プロセスモデル
     * @return true 指定されたプロセスモデルに対応する区画が存在する / false 指定されたプロセスモデルに対応する区画が存在しない
     */
    private boolean hasProcessModel(final EList<STAMPCompartment> compartments,
            final ProcessModel processModel) {
        return Compartments.getProcessModelCompartment(compartments, processModel) != null;
    }
}
