package net.astah.stpa.stamp.ui.view;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.List;
import java.util.Vector;

import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSeparator;
import javax.swing.ListCellRenderer;
import javax.swing.ListSelectionModel;

import org.eclipse.emf.edit.provider.IDisposable;

import net.astah.emf.edit.provider.ExtraLabelVariant;
import net.astah.emf.edit.provider.ItemProviders;
import net.astah.emf.edit.provider.swing.AdapterFactoryVariantLabelProvider;
import net.astah.emf.edit.provider.swing.renderer.LabelProviderListCellRenderer;
import net.astah.golf.widget.swing.ILabelProvider;
import net.astah.jomt.jsystem.PropertyManager;
import net.astah.jomt.jview.KeyDialog;
import net.astah.stpa.stamp.Element;

/**
 * エレメントをリストで単一選択するダイアログを提供します。
 */
public class SelectElementListDialog<T extends Element> extends KeyDialog
        implements ActionListener {

    private static final long serialVersionUID = 1L;
    private int resultState = CANCEL;
    private T element;
    private final List<T> elementList;

    /** OK押下時の結果コード定義 */
    public static final int OK = 1;

    /** CANCEL押下時の結果コード定義 */
    public static final int CANCEL = 2;

    /** リストのScrollPaneの幅 */
    private static final int LIST_SCROLL_PANE_WIDTH = 400;

    /** リストのScrollPaneの高さ */
    private static final int LIST_SCROLL_PANE_HEIGHT = 150;

    private final JList<T> list;
    private String description = null;
    private final JPanel northPanel;

    /**
     * ダイアログを構築します.
     * 
     * @param owner
     *            ダイアログが表示される {@code Frame}
     * @param elementList
     *            選択肢項目群
     * @param selectedElement
     *            初期選択項目
     */
    public SelectElementListDialog(final Frame owner, final List<T> elementList,
            final T selectedElement) {
        super(owner, true);
        this.elementList = elementList;

        setLocation((int) owner.getBounds().getCenterX(), (int) owner.getBounds().getCenterY());

        final Container contentPane = getContentPane();
        contentPane.setLayout(new BorderLayout());

        northPanel = new JPanel();
        northPanel.setLayout(new BoxLayout(northPanel, BoxLayout.Y_AXIS));
        contentPane.add(northPanel, BorderLayout.NORTH);

        list = createJList();
        if (selectedElement != null) {
            list.setSelectedValue(selectedElement, true);
        }
        final JScrollPane scrollPane = new JScrollPane();
        scrollPane.getViewport().setView(list);
        scrollPane.setPreferredSize(new Dimension(LIST_SCROLL_PANE_WIDTH, LIST_SCROLL_PANE_HEIGHT));
        contentPane.add(scrollPane, BorderLayout.CENTER);

        contentPane.add(createOkPanel(), BorderLayout.SOUTH);

        pack();
    }

    private Component createOkPanel() {
        final JPanel okPanel = new JPanel(new FlowLayout(FlowLayout.RIGHT));
        final JButton okButton = new JButton(PropertyManager.getProjectViewResourceBundle()
                .getString("projectview.button.ok.label"));
        okButton.setActionCommand("OK");
        okButton.addActionListener(this);
        final JButton cancelButton = new JButton(PropertyManager.getProjectViewResourceBundle()
                .getString("projectview.button.cancel.label"));
        cancelButton.setActionCommand("CANCEL");
        cancelButton.addActionListener(this);
        if (PropertyManager.isMacOS()) {
            okPanel.add(cancelButton);
            okPanel.add(okButton);
        } else {
            okPanel.add(okButton);
            okPanel.add(cancelButton);
        }

        final JPanel okBasePanel = new JPanel(new BorderLayout());
        okBasePanel.add(new JSeparator(), BorderLayout.CENTER);
        okBasePanel.add(okPanel, BorderLayout.SOUTH);
        return okBasePanel;
    }

    private JList<T> createJList() {
        final JList<T> list = new JList<>(new Vector<T>(elementList));
        list.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        list.setCellRenderer(createListCellRenderer());
        list.addPropertyChangeListener("ancestor", new PropertyChangeListener() {
            @Override
            public void propertyChange(final PropertyChangeEvent evt) {
                if (evt.getNewValue() == null) {
                    final Object cellRenderer = list.getCellRenderer();
                    if (cellRenderer instanceof IDisposable) {
                        ((IDisposable) cellRenderer).dispose();
                    }
                }
            }
        });
        return list;
    }

    /**
     * リストの{@code ListCellRenderer}を作成します。
     * 
     * @return ListCellRenderer
     */
    protected ListCellRenderer<? super T> createListCellRenderer() {
        return new LabelProviderListCellRenderer(createLabelProvider());
    }

    /**
     * リスト項目の表示値を実装したラベルプロバイダーを作成します。
     * 
     * @return ラベルプロバイダー
     */
    protected ILabelProvider createLabelProvider() {
        return new AdapterFactoryVariantLabelProvider(
                ItemProviders.getAdapterFactory(elementList.get(0).eContainer()),
                ExtraLabelVariant.MODEL_LIST);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void actionPerformed(final ActionEvent e) {
        final String action = e.getActionCommand();
        if (action.equals("OK")) {
            resultState = OK;
            element = list.getSelectedValue();
        } else if (action.equals("CANCEL")) {
            resultState = CANCEL;
        }
        setVisible(false);
    }

    /**
     * ダイアログの結果コードを返します.
     * 
     * @return {@link #OK} または {@link #CANCEL}
     */
    public int getResultState() {
        return resultState;
    }

    /**
     * OK押下時に選択されていた項目を返します.
     * 
     * @return 被選択項目
     */
    public T getSelectedElement() {
        return element;
    }

    /**
     * 説明文言を設定します.
     * 
     * @param description
     *            説明文言
     *            <p>
     *            非nullの場合に表示します.nullの場合は表示域を消去します.
     *            </p>
     */
    public void setDescription(final String description) {
        if (this.description != null) {
            northPanel.remove(0);
            pack();
        }

        if (description != null) {
            northPanel.add(createDescriptionPanel(description), 0);
            pack();
        }

        this.description = description;
    }

    private Component createDescriptionPanel(final String description) {
        final JPanel panel = new JPanel(new FlowLayout(FlowLayout.LEFT));
        final JLabel label = new JLabel();
        label.setText(description);
        panel.add(label);
        return panel;
    }

}
