package net.astah.stpa.stamp.util;

import java.util.Locale;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.uml2.common.util.UML2Util;

import com.google.common.base.Objects;

import lombok.NonNull;
import net.astah.stpa.stamp.HCFHintWordSet;
import net.astah.stpa.stamp.STPAAnalysis;
import net.astah.stpa.stamp.resource.STAMPResources;

/**
 * {@link HCFHintWordSet} 関連のユーティリティクラスです。
 */
public class HCFHintWordSets {

    /**
     * {@code hintWordSet} がプロジェクト内に取り込まれているかどうかを返します。
     * 
     * @param hintWordSet
     *            プロジェクト内に取り込まれているか判定するヒントワードセット
     * @param owner
     *            プロジェクトのヒントワードセットの親
     * @return {@code hintWordSet} がプロジェクト内に取り込まれているか否か
     */
    public static boolean isHCFHintWordSetPresent(
            @NonNull final HCFHintWordSet hintWordSet,
            @NonNull final STPAAnalysis owner) {
        return owner.getHCFHintWordSets().stream()
                .anyMatch(e -> Objects.equal(hintWordSet.getName(), e.getName())
                        && Objects.equal(hintWordSet.getLanguage(), e.getLanguage()));
    }

    /**
     * 指定されたシステム組み込みのヒントワードセットと対応付く、プロジェクト内に取り込まれているヒントワードセットを返します。
     * 
     * @param hintWordSet
     *            システム組み込みのヒントワードセット
     * @param owner
     *            プロジェクトのヒントワードセットの所有者モデル
     * @return プロジェクト内に取り込まれていたらそのヒントワードセット。取り込まれていない場合は <code>null</code>。
     */
    public static HCFHintWordSet getHintWordSetInProject(
            @NonNull final HCFHintWordSet hintWordSet,
            @NonNull final STPAAnalysis owner) {
        return owner.getHCFHintWordSets().stream()
                .filter(e -> Objects.equal(hintWordSet.getName(), e.getName())
                        && Objects.equal(hintWordSet.getLanguage(), e.getLanguage()))
                .findFirst().orElse(null);
    }

    /**
     * {@code hintWordSet} がシステム組み込みのヒントワードセットかどうかを返します。
     * 
     * @param hintWordSet
     *            システム組み込みであるか判定するヒントワードセット
     * @param owner
     *            プロジェクトのヒントワードセットの所有者モデル
     * @return アプリに保存されている場合は <code>true</code>、それ以外は <code>false</code>。
     */
    public static boolean isHCFHintWordSetInApp(
            @NonNull final HCFHintWordSet hintWordSet,
            @NonNull final STPAAnalysis owner) {
        return hintWordSet.getOwner() != owner;
    }

    /**
     * システム組み込みのヒントワードセットとプロジェクトに保存されているヒントワードセットを返します。
     * 
     * @param analysis
     *            プロジェクトのヒントワードセットの親
     * @param resourceSet
     *            システム組み込みのヒントワードセットのロード先
     * @return アプリとプロジェクトに保存されているヒントワードセット
     */
    public static EList<HCFHintWordSet> getHintWordSets(final STPAAnalysis analysis,
            final ResourceSet resourceSet) {
        final EList<HCFHintWordSet> hintWordSets = new BasicEList<HCFHintWordSet>();
        // プロジェクトヒントワードセットの取得
        hintWordSets.addAll(analysis.getHCFHintWordSets());
        // デフォルトヒントワードセットの取得
        final EList<HCFHintWordSet> defaultSets = STAMPResources
                .getLibraryDefaultHCFHintWordSets(resourceSet);
        hintWordSets.addAll(defaultSets);
        return hintWordSets;
    }

    /**
     * アプリとプロジェクトに保存されているヒントワードセットから起動言語に合うものを返します。
     * <p>
     * デフォルトと、すでに他から参照されているヒントワードセットは起動言語に関係なく含まれます。
     * </p>
     * 
     * @param analysis
     *            プロジェクトのヒントワードセットの親
     * @param resourceSet
     *            システム組み込みのヒントワードセットのロード先
     * @return アプリとプロジェクトに保存されている起動言語に合うヒントワードセット
     */
    public static EList<HCFHintWordSet> getLanguageHintWordSets(
            final STPAAnalysis analysis,
            final ResourceSet resourceSet) {
        final EList<HCFHintWordSet> hintWordSets = new BasicEList<HCFHintWordSet>();
        EList<HCFHintWordSet> allHintWordSets = getHintWordSets(analysis, resourceSet);
        for (HCFHintWordSet hintWordSet : allHintWordSets) {
            if (Objects.equal(hintWordSet.getLanguage(), Locale.getDefault().getLanguage())) {
                hintWordSets.add(hintWordSet);
            } else if (!UML2Util.getNonNavigableInverseReferences(hintWordSet).isEmpty()) {
                // 他から参照されているヒントワードセットはlanguageに関係なくリストに追加する
                hintWordSets.add(hintWordSet);
            }
        }
        // デフォルトに設定されているヒントワードセットはlanguageに関係なくリストに追加する
        HCFHintWordSet defaultHintWordSet = allHintWordSets.get(0);
        if (!hintWordSets.contains(defaultHintWordSet)) {
            hintWordSets.add(0, defaultHintWordSet);
        }
        return hintWordSets;
    }
}
